#### Load R packages
library("dplyr")
library("stringr")
library("reticulate")
library("gtools")
library("TheoChemR") ## https://github.com/clacor/TheoChemR

#### Load python package morfeus
## For an introduction see: https://rstudio.github.io/reticulate/
use_python("C:/ProgramData/Miniconda3/", required = TRUE)
mor <- import("morfeus")

#### Specifiy path to data
path <- "C:/path/to/data"

filepos <- "/logfiles/"
sdfpos <- "/sdffiles/"

#### Read DFT calculated parameters for each complex
L <- list.dirs(paste0(path, filepos), FALSE, recursive = FALSE)

xyz_tot <- list()
vib_tot <- list()
vibmot_tot <- list()
ener_tot <- list()
dipole_tot <- list()
hlgap_tot <- list()

## Takes potentially very long
for(i in 1:length(L)) {
  files <- dir(paste0(path, filepos, L[i]), pattern = ".log")
  files <- mixedsort(files)

  xyz_tot[[i]] <- vector("list", length = length(files))
  vib_tot[[i]] <- vector("list", length = length(files))
  vibmot_tot[[i]] <- vector("list", length = length(files))
  ener_tot[[i]] <- vector("list", length = length(files))
  dipole_tot[[i]] <- vector("list", length = length(files))
  hlgap_tot[[i]] <- vector("list", length = length(files))

  names(xyz_tot[[i]]) <- files
  names(vib_tot[[i]]) <- files
  names(vibmot_tot[[i]]) <- files
  names(ener_tot[[i]]) <- files
  names(dipole_tot[[i]]) <- files
  names(hlgap_tot[[i]]) <- files

  filespos <- paste0(path, filepos, L[i], "/", files)

  for(j in 1:length(filespos)) {
    temp_gauss_out <- readLines(filespos[j])
    # xyz
    xyz_tot[[i]][[j]] <- readgaussgeom(logcont = temp_gauss_out)
    # Vibrations
    vib_tot[[i]][[j]] <- readgaussout(logcont = temp_gauss_out)
    vibmot_tot[[i]][[j]] <- readgaussvibmotion(logcont = temp_gauss_out)
    # Thermodynamic energies
    ener_tot[[i]][[j]] <- readgaussener(logcont = temp_gauss_out)
    # Dipole moment
    dipole_tot[[i]][[j]] <- readgaussdipole(logcont = temp_gauss_out)
    # Frontier molecular orbital energies
    hlgap_tot[[i]][[j]] <- readgaussfroorb(logcont = temp_gauss_out)
  }
  cat(paste("Finished", i, "at", Sys.time(), "\n"))
}

xyz <- list()
vib <- list()
vibmot <- list()
ener <- list()
dipole <- list()
hlgap <- list()

for(i in 1:length(L)) {
  numbers <- unlist(lapply(str_split(names(ener_tot[[i]]), "_"), "[[", 1))
  sorted_ener <- split(ener_tot[[i]], numbers)

  ## Looking for conformer with lowest energy
  keep_conf <- unlist(lapply(sorted_ener, function(x) {
    tmp <- unlist(lapply(x, "[[", 4))
    names(x)[which.min(tmp)]
  }))

  which_to_keep <- names(ener_tot[[i]]) %in% keep_conf

  xyz[[i]] <- xyz_tot[[i]][which_to_keep]
  vib[[i]] <- vib_tot[[i]][which_to_keep]
  vibmot[[i]] <- vibmot_tot[[i]][which_to_keep]
  ener[[i]] <- ener_tot[[i]][which_to_keep]
  dipole[[i]] <- dipole_tot[[i]][which_to_keep]
  hlgap[[i]] <- hlgap_tot[[i]][which_to_keep]
}

## Load propane energy
propener <- readgaussener(paste0(path, "/Propane.log"))

#########
## Calculate descriptors via Morfeus functions

morflist <- list()

for(i in 1:length(xyz)) {

  mat <- matrix(NA, length(xyz[[i]]), 44)
  rownames(mat) <- names(xyz[[i]])[1:length(xyz[[i]])]
  colnames(mat) <- c("ConeAngle", "BiteAngle", "L", "B1", "B5", "buriedL", "buriedB1", "buriedB5", "SASA", "SASA_Rh", "SASA_Volume", "SolidAngleConeAngle", "SolidAngleG",
                     "PercBurVol", "BurVol", "DisVol", "OctVol1", "OctVol2", "OctVol3", "OctVol4", "OctVol5", "OctVol6", "OctVol7", "OctVol8", "HemiSpVolUp", "HemiSpVolDown", 
                     "COFreq", "COIR", "CORaman", "MulChaRh", "MulChaP1", "MulChaP2", "MulChaC", "MulChaCl",
                     "distRhP1", "distRhP2", "PyramP1", "PyramP2", "PyraAngleP1", "PyraAngleP2",
                     "DipoleMom", "Ehomo", "Elumo", "dEhomolumo")
  
  for(j in 1:length(xyz[[i]])) {
    testdata <- xyz[[i]][[j]]
    elements <- testdata$Sym
    coordinates <- as.matrix(testdata[, c("X", "Y", "Z")])
    whichRh <- which(elements == "Rh")
    whichP <- which(elements == "P")

    ## Exclude Cl and CO ligand from coordination sphere for some calculations
    cCl <- whichisclosest(testdata, "Cl", "Rh1")$closeLab
    cC <- whichisclosest(testdata, "C", "Rh1")$closeLab
    cO <- whichisclosest(testdata, "O", cC)$closeLab
    elementsWO <- (testdata %>% filter(!Lab %in% c(cCl, cC, cO)))$Sym
    coordinatesWO <- as.matrix((testdata %>% filter(!Lab %in% c(cCl, cC, cO)))[, c("X", "Y", "Z")])
    whichRhWO <- which(elementsWO == "Rh")
    whichPWO <- which(elementsWO == "P")
    
    ## Calculate center of mass from P atoms to create dummy atom for orientation
    dummyXYZ <- colMeans(coordinates[whichP[1:2], ])
    dummy <- rbind(coordinatesWO, dummyXYZ)
    elementsWOd <- c(elementsWO, "H")
    coordinatesWOd <- rbind(coordinatesWO, dummyXYZ, deparse.level = 0)
    whichdummy <- nrow(coordinatesWOd)

    ## Use python package morfeus to calculate different parameters
    # Cone angle
    res <- mor$ConeAngle(elementsWO, coordinatesWO, r_to_py(whichRhWO))
    mat[j, 1] <- res$cone_angle

    # Bite angle P-Rh-P
    res2 <- mor$BiteAngle(coordinates, r_to_py(whichRh), r_to_py(whichP[1]), r_to_py(whichP[2]))
    mat[j, 2] <- res2$angle

    # Sterimol
    res3 <- mor$Sterimol(elementsWOd, coordinatesWOd, r_to_py(whichRhWO), r_to_py(whichdummy))
    mat[j, 3] <- res3$L_value
    mat[j, 4] <- res3$B_1_value
    mat[j, 5] <- res3$B_5_value 

    # Buried Sterimol
    res3$bury(method="delete")
    mat[j, 6] <- res3$L_value
    mat[j, 7] <- res3$B_1_value
    mat[j, 8] <- res3$B_5_value

    # Solvent Accesable Surface Area (SASA)
    res5 <- mor$SASA(elements, coordinates)
    mat[j, 9] <- res5$area
    mat[j, 10] <- res5$atom_areas[[whichRh]]
    mat[j, 11] <- res5$volume

    # SOlid angle
    res6 <- mor$SolidAngle(elementsWO, coordinatesWO, r_to_py(whichRhWO))
    mat[j, 12] <- res6$cone_angle
    mat[j, 13] <- res6$G / 100

    # Buried volume
    ## Sometimes throws a timeout when computational power is low
    tryCatch({
      res7 <- mor$BuriedVolume(elementsWOd, coordinatesWOd, r_to_py(whichRhWO), z_axis_atoms = r_to_py(c(whichRhWO, whichdummy)), xz_plane_atoms = r_to_py(c(whichRhWO, whichPWO[1:2])))
      mat[j, 14] <- res7$fraction_buried_volume
      mat[j, 15] <- res7$buried_volume

      res7$compute_distal_volume(method="buried_volume")
      mat[j, 16] <- res7$distal_volume

      res7$octant_analysis()
      mat[j, 17:24] <- unlist(res7$octants$buried_volume)
    }, error = function(e) e)
    mat[j, 25] <- sum(mat[j, 17:20])
    mat[j, 26] <- sum(mat[j, 21:24])

    # CO Vibration
    lim_range_vibs <- subset(data.frame(t(vib[[i]][[j]])), BPosi > 1800 & BPosi < 2500)
    vib_nums <- rownames(lim_range_vibs)
    COvib <- which.max(rowSums(abs(subset(vibmot[[i]][[j]], Lab == cC & Vib %in% vib_nums)[, 6:8])) + rowSums(abs(subset(vibmot[[i]][[j]], Lab == cO & Vib %in% vib_nums)[, 6:8])))
    mat[j, 27:29] <- unlist(as.vector(lim_range_vibs[COvib, ]))

    # Mullikan charges
    mat[j, 30:34] <- filter(testdata, Lab == "Rh1" | Lab == "P1" | Lab == "P2" | Lab == cC | Lab == cCl)$MullikanCharge

    # Bond length Rh-P
    tmp <- filter(testdata, Lab == "Rh1" | Lab == "P1" | Lab == "P2") %>% arrange(desc(AtomNum))
    mat[j, 35:36] <- as.matrix(stats::dist(tmp[, 7:9]))[2:3, 1]

    # Pyramidalization
    res8 <- mor$Pyramidalization(coordinates, r_to_py(whichP[[1]]))
    res9 <- mor$Pyramidalization(coordinates, r_to_py(whichP[[2]]))
    mat[j, 37:38] <- c(res8$P, res9$P)
    mat[j, 39:40] <- c(res8$P_angle, res9$P_angle)

    # Dipole moment
    mat[j, 41] <- dipole[[i]][[j]][4]

    # HOMO-LUMO energy gap
    mat[j, 42:44] <- hlgap[[i]][[j]]
  }
  morflist[[i]] <- mat
  cat(paste0("Finished with ", i, " at ", Sys.time(), "\n"))
}

#### Include xTB
## First download statically linked xtb library from https://github.com/grimme-lab/xtb/releases/latest
## unpack,
## include xtb/bin/ and xtb/share/xtb in PATH variable

## Make xyz files from output geometry
for(j in 1:length(L)) {
  if(!dir.exists(paste0(path, filepos, L[j], "/xyz"))) {
    dir.create(paste0(path, filepos, L[j], "/xyz"))
  }
  for(i in 1:length(xyz[[1]])) {
    write.table(rbind(c(nrow(xyz[[j]][[i]]), rep(" ", 3)), rep(" ", 4), xyz[[j]][[i]][, c("Sym", "X", "Y", "Z")]), file = paste0(path, filepos, L[j], "/xyz/", names(xyz[[j]])[i], ".xyz"),
      row.names = FALSE, col.names = FALSE, quote = FALSE, sep = "\t")
  }
}

fukuilist <- list()
GEIlist <- list()

## Use xyz coordinates for xTB calculations
for(j in 1:length(L)) {
  fukuilist[[j]] <- list()
  GEIlist[[j]] <- list()
  files2 <- dir(paste0(path, filepos, L[j],"/xyz"), pattern = ".xyz")
  files2 <- mixedsort(files2)
  
  for(i in 1:length(files2)) {
    # Fukui
    tmp <- system2("xtb", paste0("\"", gsub("/", "\\\\", paste0(path, filepos, L[j],"/xyz/", files2[i], "\" --vfukui"))), stderr=NULL, stdout=TRUE)
    beginn <- which(tmp == "     #        f(+)     f(-)     f(0)")+1
    ending <- which(tmp == "          |                Property Printout                |")-2
    temp <- str_split(paste(tmp[beginn:ending], collapse = ""), "\\s+")[[1]][-1]
    atoms <- temp[seq(1, length(temp), 4)]
    num <- gsub("[^[:digit:]]", "", atoms)
    sym <- gsub("[[:digit:]]", "", atoms)
    fukui <- temp[-seq(1, length(temp), 4)]
    fukuilist[[j]][[i]] <- data.frame(Num = num, Sym = sym, fp = as.numeric(fukui[seq(1, length(fukui), 3)]),
      fm = as.numeric(fukui[seq(2, length(fukui), 3)]),
      f0 = as.numeric(fukui[seq(3, length(fukui), 3)]))
    
    # Electrophilicity, Nucleophilicity, Electrofugality, Nucleofugality
    tmp <- system2("xtb", paste0("\"", gsub("/", "\\\\", paste0(path, filepos, L[j],"/xyz/", files2[i], "\" --vomega"))), stderr=NULL, stdout=TRUE)
    ElePhil <- sum(as.numeric(str_match(tmp, "Global electrophilicity index \\(eV\\):\\s+(\\d+\\.\\d+)")[, 2]), na.rm = TRUE)
    IP <- sum(as.numeric(str_match(tmp, "delta SCC IP \\(eV\\):\\s+(\\d+\\.\\d+)")[, 2]), na.rm = TRUE)
    EA <- sum(as.numeric(str_match(tmp, "delta SCC EA \\(eV\\):\\s+(\\d+\\.\\d+)")[, 2]), na.rm = TRUE)
    NucPhil <- -IP
    EleFug <- (IP-3*EA)^2/(8*(IP-EA))
    NucFug <- (3*IP-EA)^2/(8*(IP-EA))
    GEIlist[[j]][[i]] <- data.frame(IP, EA, ElePhil, NucPhil, EleFug, NucFug)

    if(i %% 100 == 0) {
      cat(paste("Finished", j, ",", i, "at", Sys.time(), "\n"))
    }
  }
  names(fukuilist[[j]]) <- files2
  names(GEIlist[[j]]) <- files2
}

#### Sanity check
identical(gsub("\\.xyz", "", names(fukuilist[[1]])), rownames(morflist[[1]]))

## Append Fukui and GEI data to morfeus parameter list
oldmorflist <- morflist

for(i in 1:length(morflist)) {
  fukuimat <- do.call(rbind, lapply(fukuilist[[i]], function(x) {as.vector(t(subset(x, Sym == "Rh" | Sym == "P")[, 3:5]))}))
  colnames(fukuimat) <- c("fukuip_Rh", "fukuim_Rh", "fukui0_Rh", "fukuip_P1", "fukuim_P1", "fukui0_P1", "fukuip_P2", "fukuim_P2", "fukui0_P2")
  GEImat <- do.call(rbind, GEIlist[[i]])
  morflist[[i]] <- cbind(oldmorflist[[i]], fukuimat, GEImat)
}

#### Collapse energy vector and calculate energy difference
e4 <- do.call(rbind, ener[[1]])
e6 <- do.call(rbind, ener[[2]])

## Test if data order is identical in 4- and 6-times coordinated case
all(unlist(lapply(stringr::str_split(rownames(e4), "_"), "[[", 1)) ==
  unlist(lapply(stringr::str_split(rownames(e6), "_"), "[[", 1)))

ediff <- e6[, 4] - (e4[, 4] + propener[4])

## Combine dE with parameters
## also remove NAs from morfeus calculation
fin_df4 <- na.omit(data.frame(morflist[[1]], dE = ediff)) 
fin_df6 <- na.omit(data.frame(morflist[[2]], dE = ediff))

#### Reorder charges and bond length so that P1 is the atom with lower mullikan charge
tmp <- t(apply(fin_df4[, c("MulChaP1", "MulChaP2")], 1, order))
for(i in 1:nrow(tmp)) {
  fin_df4[i, c("MulChaP1", "MulChaP2")] <- fin_df4[i, c("MulChaP1", "MulChaP2")][tmp[i, ]]
  fin_df4[i, c("distRhP1", "distRhP2")] <- fin_df4[i, c("distRhP1", "distRhP2")][tmp[i, ]]
  fin_df4[i, c("PyramP1", "PyramP2")] <- fin_df4[i, c("PyramP1", "PyramP2")][tmp[i, ]]
  fin_df4[i, c("fukuip_P1", "fukuip_P2")] <- fin_df4[i, c("fukuip_P1", "fukuip_P2")][tmp[i, ]]
  fin_df4[i, c("fukuim_P1", "fukuim_P2")] <- fin_df4[i, c("fukuim_P1", "fukuim_P2")][tmp[i, ]]
  fin_df4[i, c("fukui0_P1", "fukui0_P2")] <- fin_df4[i, c("fukui0_P1", "fukui0_P2")][tmp[i, ]]
}

tmp <- t(apply(fin_df6[, c("MulChaP1", "MulChaP2")], 1, order))
for(i in 1:nrow(tmp)) {
  fin_df6[i, c("MulChaP1", "MulChaP2")] <- fin_df6[i, c("MulChaP1", "MulChaP2")][tmp[i, ]]
  fin_df6[i, c("distRhP1", "distRhP2")] <- fin_df6[i, c("distRhP1", "distRhP2")][tmp[i, ]]
  fin_df6[i, c("PyramP1", "PyramP2")] <- fin_df6[i, c("PyramP1", "PyramP2")][tmp[i, ]]
  fin_df6[i, c("PyraAngleP1", "PyraAngleP2")] <- fin_df6[i, c("PyraAngleP1", "PyraAngleP2")][tmp[i, ]]
  fin_df6[i, c("fukuip_P1", "fukuip_P2")] <- fin_df6[i, c("fukuip_P1", "fukuip_P2")][tmp[i, ]]
  fin_df6[i, c("fukuim_P1", "fukuim_P2")] <- fin_df6[i, c("fukuim_P1", "fukuim_P2")][tmp[i, ]]
  fin_df6[i, c("fukui0_P1", "fukui0_P2")] <- fin_df6[i, c("fukui0_P1", "fukui0_P2")][tmp[i, ]]
}

#### Save dataset
# save(findf4, findf6, file = paste0(path, "/MorfeusParamsBisPhos_", Sys.Date(), ".rda"))



